/*
 * vf.c version 2.0 part of the vfloppy 2.0 package
 *
 * This version delivered in 2010 by Fred Jan Kraan (fjkraan@xs4all.nl)
 *
 * vf is placed under the GNU General Public License in March 2010.
 *
 *  This file is part of Vfloppy 2.0.
 *
 *  Vfloppy is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Vfloppy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Vfloppy; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * vf.c - functions for vformat, vfread, vfwrite
 */

#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#include "vf.h"

extern int debug;

void parseArgs (int argc, char **argv, argData_t *pArgData)
{
	int index, c;
	pArgData->imageName = NULL;
	pArgData->fileName = NULL;
	pArgData->debug = 0;
	
	opterr = 0;
     
	while ((c = getopt (argc, argv, "d:")) != -1)
	{
		switch (c)
		{
		case 'd':
			pArgData->debug = atoi(optarg);
			break;
		case '?':
			if (optopt == 'd')
				fprintf (stderr, "Option -%c requires an argument.\n", optopt);
			else if (isprint (optopt))
				fprintf (stderr, "Unknown option `-%c'.\n", optopt);
			else
				fprintf (stderr, "Unknown option character `\\x%x'.\n", optopt);
			return;
		default:
			abort ();
		}
	}
	
	c = 0;
	for (index = optind; index < argc; index++)
	{
		if (pArgData->debug > 1) printf (" parseArgs; Arg name %d: %s\n", index, argv[index]);
		if (c == 0)
		{
			pArgData->imageName = argv[index];
		}
		if (c == 1)
		{
			pArgData->fileName = argv[index];
		}
		c++;
	}	
	if (pArgData->debug > 1)
	{
		printf(" parseArgs; image name: %s\n", pArgData->imageName);
		printf(" parseArgs; file name: %s\n", pArgData->fileName);
		printf(" parseArgs; debug: %d\n", pArgData->debug);
	}
}


void printDirectory(FILE *imageFile)
{
//	imageFile = fopen("../px8_epsntf20.d88", "rb");
	int i, fileSize;
	long int extendLocation;
	unsigned char extend[EXTEND];
	unsigned char fileName[13];
	fileName[8] = '.';
	fileName[12] = 0;
	for (i = 0; i < DIRECTORYEXTENDS; i++)
	{
		extendLocation = getDirExtendLocation(i);
		if (debug > 2) printf(" printDirectory; reading at %lX\n", extendLocation);
		
		readDisk((char *)extend, extendLocation, EXTEND, imageFile);
		if (extend[0] != FORMATPATTERN) 
		{
			memcpy(&fileName[0], &extend[1], 8);
			memcpy(&fileName[9], &extend[9], 3);
			fileName[READONLYBIT]   &= 0x7F;
			fileName[SYSTEMFILEBIT] &= 0x7F;
			fileName[ARCHIVEBIT]    &= 0x7F;
			fileSize = extend[EXTENDRECORDS] * RECORD + (extend[EXTENDNO] & 1) * (1<<14);
			printf("%d: %s %c%c%c %d %d\n", 
			       extend[0], 
			       fileName, 
			       setCharOnHibit(extend[READONLYBIT],   'r'), 
			       setCharOnHibit(extend[SYSTEMFILEBIT], 's'), 
			       setCharOnHibit(extend[ARCHIVEBIT],    'a'), 
			       extend[EXTENDNO] / 2, 
			       fileSize );
		}
	}
}

char setCharOnHibit(unsigned char testChar, char replaceChar)
{
	return (testChar > 0x7F) ? replaceChar : ' ';
}

void convertFileName(fileData_t *pFileData)
{
	// This is copied from the old code, but in need of serious refactoring. 
	// To be combined with the support for files from other than the local directory
	
	/* conversie fileName dirFileName
	   - to uppercase
	   - add spaces to basename to 8 chars
	   - remove dot
	   - add spaces to extension to 3 chars
	 */
	int i = 0;
	int j = 0;
	int c;
	for (c = 0; c < 11; c++)
		pFileData->cpmFileName[c] = ' ';
	pFileData->cpmFileName[11] = '\0';

	/* walk though the array to fill and check the
	 * input array:
	 * - end of string: 	exit
	 * - a dot:		write the rest at the extension location
	 * Most problems are not checked here, but will hopefully fail
	 * nicely at filename comparison.
	 */
	for (i = 0; i < 12; i++) 
	{
		if (pFileData->fileName[i] == '\0')	// stop on end of string
			break;
		if (pFileData->fileName[i] == '.') 	// skip to filename extension
		{
			j = 8;		// start filename extension
		} else {
			if (j < 11)	// don't copy the end of string
			{
				pFileData->cpmFileName[j] = toupper(pFileData->fileName[i]);
				j++;
			}	
		}
	} 

	if (debug > 1) printf(" convertFileName; filename: \'%s\'\n", pFileData->cpmFileName); 
}

void findFileExtends(FILE *imageFile, fileData_t *pFileData)
{
/** findFileExtends - searches for a filename in the image and writes the used
  *            extend numbers to the struct extends[] and sets the found flag.
  */
	int i;
	long int extendLocation;
	unsigned char extend[EXTEND];
	unsigned char foundFilename[11];
	foundFilename[11] = '\0';
	for (i = 0; i < DIRECTORYEXTENDS; i++)  // iterate through directory entries
	{
		extendLocation = getDirExtendLocation(i);
		if (debug > 2) printf(" findFileExtends; reading at %lX\n", extendLocation);
		
		readDisk((char *)extend, extendLocation, EXTEND, imageFile);	
		
		if (extend[0] != FORMATPATTERN) {
			memcpy(&foundFilename, &extend[FILENAMEBASE], 11);
			stripHighbit(foundFilename);
			if (strcmp((const char *)foundFilename, (const char *)pFileData->cpmFileName) == 0) 
			{
				pFileData->extends[extend[EXTENDNO]] = i;
				if (debug > 2) printf(" findFileExtends; found %s at %X extend number %X\n", foundFilename, i, extend[i]);
				pFileData->found = 1;
			}
		}
 	}
		
	if (debug > 1) printf(" findFileExtends; found extends ");
	{
		if (debug > 1 && pFileData->extends[i] != -1) printf("%X: %X, ", i, pFileData->extends[i]);
	}
	if (debug > 1) printf("\n");
}

void stripHighbit(unsigned char *fileName)
{
			fileName[READONLYBIT]   = 0x7f & fileName[READONLYBIT];	/* clear Read only bit */
			fileName[SYSTEMFILEBIT] = 0x7f & fileName[SYSTEMFILEBIT];	/* clear System file bit */
			fileName[ARCHIVEBIT]    = 0x7f & fileName[ARCHIVEBIT];	/* clear Archive bit */ 
}

void findFileBlocks(FILE *imageFile, fileData_t *pFileData)
{
	int i;
	for (i = 0; i < DIRECTORYEXTENDS; i++)
	{
		if (pFileData->extends[i] != -1) 
		{
			if (debug > 1) printf(" findFileBlocks; copying from extend %X\n", pFileData->extends[i]);
			copyBlocks(imageFile, pFileData, pFileData->extends[i]);
		}
	}
}

int checkFileExistance(char *fileName)
{
	FILE *file = fopen(fileName, "rb");
	if (file == NULL)
	{
		if (debug > 1) 
			printf(" checkFileExistance; file %s does not exist\n", fileName);
		return 0;
	}
	fclose(file);
	if (debug > 1) 
		printf(" checkFileExistance; file %s does exist\n", fileName);
	return 1;
}

void copyBlocks(FILE *imageFile, fileData_t *pFileData, int extendNumber)
{
	int i, j, block;
	long int extendLocation;
	unsigned char extend[EXTEND];
	
	extendLocation = getDirExtendLocation(extendNumber);
	if (debug > 2) printf("  copyBlocks; reading at %lX, ", extendLocation);
	
	readDisk((char *)extend, extendLocation, EXTEND, imageFile);	
	
	for (j = 0; j < EXTBLOCKS; j++) 
	{
		block = extend[EXTBLOCKBASE + j];
		if (block != 0) storeBlockNumber(pFileData, block);
	}
	// set for every extend in correct order, so last extend will remain
	pFileData->recordsInLastExtend = extend[EXTENDRECORDS]; 
	
	if (debug > 2) {
		printf("  copyBlocks; found blocks: ");
		for (i = 0; i < BLOCKS_ON_DISK; i++) 
		{
			if (pFileData->fileBlocks[i] == 0) break;
			printf("%X, ", pFileData->fileBlocks[i]);
		}
		printf("\n");
		printf("  copyBlocks; records in last extend: %X, in last block: %X\n", 
		       pFileData->recordsInLastExtend, 
		       (pFileData->recordsInLastExtend & 0x0F) == 0 ? 0x10 : pFileData->recordsInLastExtend & 0x0F);
	}
}

void storeBlockNumber(fileData_t *pFileData, int block)
{
	int i = 0;
	while (pFileData->fileBlocks[i] != 0) i++;	// Find first empty location
	pFileData->fileBlocks[i] = block;
	if (debug > 3) printf("   storeBlockNumber; copy block: %X to fileBlocks[%X]\n", block, i);
}

void createFile(FILE *imageFile, fileData_t *pFileData)
{
	pFileData->file = fopen((const char *)pFileData->fileName, "wb");
  	if (pFileData->file == NULL) 
	{
		fprintf(stderr,"failed to open %s for writing\n", pFileData->fileName);
		exit(2);
	} 
	if (debug > 1)  printf(" createFile; opened %s for writing\n", pFileData->fileName);
	int i;
	for (i = 0; i < BLOCKS_ON_DISK; i++) // block 0 is the directory
	{
		if (pFileData->fileBlocks[i] == 0) 
		{
			if (debug > 2) printf(" createFile; position %X is 0, finished\n", i);
			break;	// end of blocks
		}
		if (debug > 1) printf(" createFile; position %X has block %X\n", i,  pFileData->fileBlocks[i]);
		if (pFileData->fileBlocks[i + 1] == 0)
		{
			lastBlockCopy(imageFile, pFileData, i);
		} 
		else 
		{
			otherBlockCopy(imageFile, pFileData, i);
		}
	}
	fclose(pFileData->file);
}

void lastBlockCopy(FILE *imageFile, fileData_t *pFileData, int block)
{
	long int recordAddress;
	int i;
	unsigned char record[RECORD];
	int recordsInLastBlock = 
		(pFileData->recordsInLastExtend & 0x0F) == 0 ? 0x10 : pFileData->recordsInLastExtend & 0x0F;
	if (debug > 2)  printf("  lastBlockCopy; %X records to copy\n", recordsInLastBlock);
	for (i = 0; i < recordsInLastBlock; i++)
	{
		recordAddress = getRecordLocation(pFileData->fileBlocks[block], i);
		if (debug > 2)  printf("  lastBlockCopy; record %X at %lX\n", i, recordAddress);
		
		readDisk((char *)record, recordAddress, RECORD, imageFile);	

		writeDisk((char *)record, 0, RECORD, pFileData->file);
	}	
}

void otherBlockCopy(FILE *imageFile, fileData_t *pFileData, int block)
{
	long int sectorAddress;
	int i;
	unsigned char sector[SECTOR];
	for (i = 0; i < (BLOCK / SECTOR); i++)
	{
		sectorAddress = getSectorLocation(pFileData->fileBlocks[block], i);
		if (debug > 2)  printf("  otherBlockCopy; sector %X at %lX\n", i, sectorAddress);
		
		readDisk((char *)sector, sectorAddress, SECTOR, imageFile);	
		
		writeDisk((char *)sector, 0, SECTOR, pFileData->file);
	}
}

long int getDirExtendLocation (int count) 
{
	return (SECTORBASE + SECTORHEADER + (DIRTRACK * TRACK * HEADS)) + 	// base offset
		count * EXTEND +					// per extend offset
		(count / (SECTOR / EXTEND)) * SECTORHEADER;		// per sector offset
}

long int getBlockLocation(int block)
{
	return BLOCKBASE + 
		block * SECTORSPERBLOCK * (SECTORHEADER + SECTOR);
} 

long int getSectorLocation(int block, int sector)
{
	return BLOCKBASE + 
		block * (BLOCK / SECTOR) * (SECTORHEADER + SECTOR) +
		sector * (SECTORHEADER + SECTOR) +
		SECTORHEADER;
}

long int getRecordLocation(int block, int record)
{
	return BLOCKBASE + 
		block * (BLOCK / SECTOR) * (SECTORHEADER + SECTOR) +
		record * (RECORD) +
		SECTORHEADER + SECTORHEADER * (record / (SECTOR / RECORD));
}

void fileDataInit(fileData_t *pFileData)
{
	pFileData->cpmFileName[11] = '\0';
	pFileData->found = 0;
	int i;
	for (i = 0; i < BLOCKS_ON_DISK; i++) pFileData->fileBlocks[i] = 0;
	for (i = 0; i < DIRECTORYEXTENDS; i++) pFileData->extends[i] = -1;	
}

int getFreeBlockCount(FILE *imageFile)
{
	int i, j, blockCount = 0;
	long int extendLocation;
	unsigned char extend[EXTEND];
	for (i = 0; i < DIRECTORYEXTENDS; i++)
	{
		extendLocation = getDirExtendLocation(i);
		if (debug > 2) printf("  getFreeBlockCount; reading at %X: %lX,", i, extendLocation);
	
		readDisk((char *)extend, extendLocation, EXTEND, imageFile);	
		if (debug > 2) printf(" extend[STATUS]: %X\n", extend[STATUS] );
		if (extend[STATUS] != FORMATPATTERN) {
			for (j = 0; j < EXTBLOCKS; j++) 
			{
				if (extend[EXTBLOCKBASE + j] != 0) blockCount++;
			}
			if (debug > 2) printf("  getFreeBlockCount; used blockCount: %d\n", blockCount);
		}
	}
	if (debug > 2) printf("  getFreeBlockCount; found %d used blocks\n", blockCount);
	return BLOCKS_ON_DISK - blockCount;
}

int getSizeInRecords(int size)
{
	return (size + RECORD - 1) / RECORD;
}

int getImageFileBlockCount(fileData_t *pFileData)
{
	int i, blockCount = 0;
	for (i = 0; i < BLOCKS_ON_DISK; i++) 
	{
		if (pFileData->fileBlocks[i] != 0) blockCount++;	
	}
	if (debug > 1) printf("  getImageFileBlockCount; checked %X blocks, %X counted\n", i, blockCount);
	return blockCount;
}

int getImageFileExtendCount(fileData_t *pFileData)
{
	int i, extendCount = 0;
	for (i = 0; i < DIRECTORYEXTENDS; i++)
	{
		if (pFileData->extends[i] != -1) extendCount++;
	}
	if (debug > 1) printf("  getImageFileExtendCount; checked %X extends, %X counted\n", i, extendCount);
	return extendCount;
}

unsigned int getFileSize(fileData_t *pFileData)
{
	int result, lSize = 0;
	FILE *file = fopen((char *)pFileData->fileName, "rb");
	if (file == NULL) 
	{
		fprintf(stderr," failed to open %s\n", pFileData->fileName);
		exit(1);
	} else {
		result = fseek (file , 0 , SEEK_END);
		lSize = ftell (file);
		if (debug > 2) printf("  getFileSize; returned: %d. File is %d bytes (0x%X)\n", result, lSize, lSize);
		fclose(file);
	}
	return lSize;
}

int getSizeInBlocks(int size)
{
	return (size + (BLOCK - 1)) / BLOCK;
}

int getSizeInExtends(int size)
{
	return (size + (EXTBLOCKS * BLOCK) - 1) / (EXTBLOCKS * BLOCK);
}

int getFreeDirectorySpace(FILE *imageFile)
{
	int i, freeExtendCount = 0;
	long int extendLocation;
	unsigned char extend[EXTEND];
	for (i = 0; i < DIRECTORYEXTENDS; i++)
	{
		extendLocation = getDirExtendLocation(i);
		if (debug > 2) printf("  getFreeDirectorySpace; reading at %X: %lX\n", i, extendLocation);
	
		readDisk((char *)extend, extendLocation, EXTEND, imageFile);	
		
		if (extend[STATUS] == FORMATPATTERN) freeExtendCount++;

	}
	if (debug > 1) printf(" getFreeDirectorySpace; %X extends free\n", 
		freeExtendCount);
	return freeExtendCount;
}

void eraseFile(FILE *imageFile, fileData_t *pFileData)
{
	int i;
	long int extendLocation;
	unsigned char extend[EXTEND];
	
	for (i = 0; i < DIRECTORYEXTENDS; i++)
	{
		if (pFileData->extends[i] != -1) {
			extendLocation = getDirExtendLocation(pFileData->extends[i]);
			if (debug > 2) 
				printf(" eraseFile; erasing extend %X: %X at %lX\n", i, pFileData->extends[i], extendLocation);

			readDisk((char *)extend, extendLocation, EXTEND, imageFile);
				
			extend[0] = FORMATPATTERN;
				
			writeDisk((char *)extend, extendLocation, EXTEND, imageFile);			
		}
	}
	if (debug > 2) printf(" eraseFile; end\n");
}

void findUnusedBlocksExtends(FILE *imageFile, fileData_t *pFileData)
{
	int i,  j;
	long int extendLocation;
	unsigned char extend[EXTEND];
	// start with making all blocks available
	for (i = 1; i < BLOCKS_ON_DISK; i++) pFileData->fileBlocks[i] = 0;
	for (i = 0; i < DIRECTORYEXTENDS; i++)
	{
		extendLocation = getDirExtendLocation(i);
		readDisk((char *)extend, extendLocation, EXTEND, imageFile);
		if (debug > 2) 
			printf(" findUnusedBlocksExtends; checking extend %X: at %lX (STATUS=%X)\n", 
			       i,  extendLocation, extend[STATUS]);
		if (extend[STATUS] != FORMATPATTERN) {
			if (debug > 2) 
				printf(" findUnusedBlocksExtends; found used extend at %X: ", i);
			for (j = 0; j < EXTBLOCKS; j++) 
			{
				if (extend[EXTBLOCKBASE + j] != 0) 
				{
					pFileData->fileBlocks[extend[EXTBLOCKBASE + j]] = extend[EXTBLOCKBASE + j];
					if (debug > 2) printf(" %X:,", extend[EXTBLOCKBASE + j]);
				}
			}
			pFileData->extends[i] = i;
			if (debug > 2) printf("\n");
		} else {
			if (debug > 2) printf(" findUnusedBlocksExtends; free extend:%X\n", i);
			pFileData->extends[i] = -1;
		}
	}
	if (debug > 2) 
	{
		printf(" findUnusedBlocksExtends; blocks found to be free: ");
		for (i = 1; i < BLOCKS_ON_DISK; i++) 
		{
			if (pFileData->fileBlocks[i] == 0)
				printf(" %X,", i);
		}
		printf("\n");
		printf(" findUnusedBlocksExtends; extends found to be free: ");
		for (i = 1; i < DIRECTORYEXTENDS; i++) 
		{
			if (pFileData->extends[i] == -1)
				printf(" %X,", i);
		}
		printf("\n");
	}
}

int getNextFreeBlock(fileData_t *pFileData, int index)
{
	int i = index + 1;
	while (pFileData->fileBlocks[i] != 0) i++;
	return i;
}

int getNextFreeExtend(fileData_t *pFileData, int index)
{
	int i = index + 1;
	while (pFileData->extends[i] != -1) i++;
	return i;
}

void writeDirectory(FILE *imageFile, fileData_t *pFileData)
{
	int i, blocksThisExtend, extCount, blkCount, fileInExtends, fileInBlocks, lastExtend = 0;
	int blockIndex = 0, extendIndex = -1;	// blocks should start at 1, extends at 0
	long int extendLocation;
	unsigned char extend[EXTEND];
	
	int remainingSize = pFileData->fileSize;
	fileInExtends = getSizeInExtends(pFileData->fileSize);
	fileInBlocks = getSizeInBlocks(pFileData->fileSize);
	if (debug > 2) printf(" writeDirectory; fileInBlocks: %X, remainingSize: %X\n", 
		fileInBlocks, remainingSize);
	for (extCount = 0; extCount < fileInExtends; extCount++)
	{
		memset(extend, 0x0, EXTEND);
		for (i = 0; i < 11; i++) extend[FILENAMEBASE + i] = pFileData->cpmFileName[i];
		
		if (extCount + 1 == fileInExtends) lastExtend = 1;
		extend[EXTENDNO] = extCount * 2; // use bits 1 and up
		extendIndex = getNextFreeExtend(pFileData, extendIndex);
		if (debug > 2) printf(" writeDirectory; extCount=%X, extIdx=%X:", extCount, extendIndex);
		
		extendLocation = getDirExtendLocation(extendIndex);
		
		if (getSizeInBlocks(remainingSize) > EXTBLOCKS)
		{
			blocksThisExtend = EXTBLOCKS;
		} 
		else 
		{
			blocksThisExtend = getSizeInBlocks(remainingSize);
		}
		if (debug > 2) printf(" bTE: %X,", blocksThisExtend);
		for (blkCount = 0; blkCount < blocksThisExtend; blkCount++)
		{
			blockIndex = getNextFreeBlock(pFileData, blockIndex);
			extend[EXTBLOCKBASE + blkCount] = blockIndex;
			if (debug > 2) printf(" blkCount=%X, blkIdx=%X ", blkCount, blockIndex);
//			if (blkCount >= EXTBLOCKS) break;
		}
		if (debug > 2) printf("\n");
		if (lastExtend) {
			extend[EXTENDRECORDS] = getSizeInRecords(remainingSize) & 0x7F;
			if (extend[EXTENDRECORDS] == 0) extend[EXTENDRECORDS] = 0x80;
			
			if (getSizeInRecords(remainingSize) > 0x7F) extend[EXTENDNO] = extend[EXTENDNO] + 1;
			if (debug > 2) printf(" writeDirectory; (last) remainingSize = 0x%X, fileInRecords = %X, extend[EXTENDRECORDS] = %X, extend[EXTENDNO] = %X\n",
			       remainingSize, getSizeInRecords(remainingSize), extend[EXTENDRECORDS], extend[EXTENDNO]);
			remainingSize = 0;
		} else {
			extend[EXTENDRECORDS] = 0x80;			// any non-last extend is full.
			extend[EXTENDNO] = extend[EXTENDNO] + 1;	// full = 256 records
			if (debug > 2) printf(" writeDirectory; remainingSize = 0x%X, fileInRecords = %X, extend[EXTENDRECORDS] = %X, extend[EXTENDNO] = %X\n",
			       remainingSize, getSizeInRecords(remainingSize), extend[EXTENDRECORDS], extend[EXTENDNO]);
			remainingSize -= EXTBLOCKS * BLOCK;
		}

		if (debug > 2) printf(" writeDirectory; location: %lX, updated remaining size: 0x%X\n", extendLocation, remainingSize);
		writeDisk((char *)extend, extendLocation, EXTEND, imageFile);
	}
}

void writeFile(FILE *imageFile, fileData_t *pFileData)
{
	int blkCount, secCount, blockIndex = 0;
	unsigned char buffer[SECTOR];
	unsigned long int sectorLocation, fileSectorLocation = 0;
	for (blkCount = 0; blkCount < getSizeInBlocks(pFileData->fileSize); blkCount++)
	{
		blockIndex = getNextFreeBlock(pFileData, blockIndex);
		for (secCount = 0; secCount < SECTORSPERBLOCK; secCount++)
		{
			sectorLocation = getSectorLocation(blockIndex, secCount);
			if (debug > 2) printf("blkCount: %X, blockIndex %X, secCount: %X, sectorLocation: %lX\n",
			       blkCount, blockIndex, secCount, sectorLocation);
			memset(buffer, FORMATPATTERN, SECTOR);
			readDisk((char *)buffer, fileSectorLocation, SECTOR, pFileData->file);
			fileSectorLocation += SECTOR;
			writeDisk((char *)buffer, sectorLocation, SECTOR, imageFile);
		}
	}
}

void initImageHeader(d88Header_t *pD88Header)
{
	int i;
	for (i = 0; i < 9; i++)	pD88Header->rsrv[i] = 0;
	pD88Header->protect = 0;
	pD88Header->type = 0;
	pD88Header->size = TRACK * HEADS * TRACKS + SECTORBASE;
	for (i = 0; i < TRACKS * HEADS; i++)   pD88Header->trkptr[i] = i * TRACK + SECTORBASE;
	for (i = TRACKS * HEADS; i < 164; i++) pD88Header->trkptr[i] = 0;
}

void createImage(char *imageName, d88Header_t *pD88Header)
{
	// sector header
	struct d88sct_t sectorHeader;
	sectorHeader.c      = 0;	// cylinder,track 0 - 39
	sectorHeader.h      = 0;	// head 0 - 1
	sectorHeader.r      = 0;	// sector 1 - 16
	sectorHeader.n      = 1;	// ?
	
	sectorHeader.nsec    = 0x1000;
	sectorHeader.dens    = 0;
	sectorHeader.del     = 0;
	
	sectorHeader.stat    = 0;
	sectorHeader.rsrv[0] = 0;
	sectorHeader.rsrv[1] = 0;
	sectorHeader.rsrv[2] = 0;
	
	sectorHeader.rsrv[3] = 0;
	sectorHeader.rsrv[4] = 0;
	sectorHeader.size    = 0x0100;

	// Fill the sector with format pattern
	unsigned char sectorBuffer[SECTOR];
	memset(sectorBuffer, FORMATPATTERN, SECTOR);
	
	if (debug > 1) printf(" createImage; opening image %s\n", imageName);
	FILE *imageFile = fopen(imageName, "wb");
	if (imageFile == NULL) 
	{
		fprintf(stderr," failed to open %s\n", imageName);
		exit(2);
	} 

	// Filling the image
	unsigned char sector, head, track;
	if (debug > 1) printf(" createImage; Creating a disk image with %d tracks, %d sides, %d sectors of %d bytes\n",
		TRACKS, HEADS, SECTORSPERTRACK, SECTOR);
	fwrite(pD88Header, SECTORBASE, 1, imageFile);
	for (track = 0; track < TRACKS; track++)
	{
		sectorHeader.c = track;
		for (head = 0; head < HEADS; head++)
		{
			sectorHeader.h = head;
			for (sector = 1; sector <= SECTORSPERTRACK; sector++)
			{
				sectorHeader.r = sector;
				fwrite(&sectorHeader, SECTORHEADER, 1, imageFile);
				fwrite(sectorBuffer, SECTOR, 1, imageFile);
			}
		}
	} 
}

FILE *openFileRO(char *imageName)
{
	FILE *imageFile = fopen(imageName, "rb");
	if (imageFile == NULL) 
	{
		fprintf(stderr," failed to open %s\n", imageName);
		exit(1);
	} else {
		int result = fseek (imageFile , 0 , SEEK_END);
		int lSize = ftell (imageFile);
		if (debug > 1) printf(" openImageRO; returned: %d. File is %d bytes (0x%X)\n", result, lSize, lSize);
	}
	return imageFile;
}

FILE *openFileRW(char *imageName)
{
	int result, lSize;
	FILE *imageFile = fopen(imageName, "r+b");
	if (imageFile == NULL) 
	{
		fprintf(stderr," failed to open %s\n", imageName);
		exit(1);
	} else {
		result = fseek (imageFile , 0 , SEEK_END);
		lSize  = ftell (imageFile);
		if (debug > 1) printf(" openFileRW; returned: %d. File is %d bytes (0x%X)\n", result, lSize, lSize);
	}
	return imageFile;
}

void readDisk(char *buffer, long int location, unsigned int size, FILE *readFile)
{
//	printf("readDisk; readDisk: %p, buffer: %p\n", readFile, buffer);
	int result;
	if (debug > 3) printf("  readDisk; reading 0x%X bytes", size);
		
	if (debug > 3) printf(" at %lX,", location);
	result = fseek(readFile, location, SEEK_SET);
	if (debug > 3) printf(" fseek returned: %d,", result);
	
	result = fread(buffer, size, 1, readFile);
	if (debug > 3) printf(" fread returned: %d\n", result);	
}

void writeDisk(char *buffer, long int location, unsigned int size, FILE *writeFile)
{
	int result;
	if (debug > 3) printf("  writeDisk; writing 0x%X bytes,", size);
			
	if (location != 0)	// file writes don't need fseek for append only files
	{
		if (debug > 3) printf(" at %lX,", location);
		result = fseek(writeFile, location, SEEK_SET);
		if (debug > 3) printf(" fseek returned: %d,", result);
	}
	
	result = fwrite(buffer, size, 1, writeFile);
	if (debug > 3) printf(" fwrite returned: %d\n", result);
}